#ifndef _KTHREAD_H
#define _KTHREAD_H
#include <linux/config.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/tqueue.h>
#include <linux/wait.h>

#include <asm/unistd.h>
#include <asm/semaphore.h>

// a structure to store all information we need	for our thread 
typedef struct kthread_struct
{
    // private data 
    struct task_struct *thread;							// Linux task structure of thread 
    struct tq_struct tq;								// Task queue need to launch thread 
    void (*function) (struct kthread_struct *kthread);	// function to be started as thread 
    struct semaphore startstop_sem;	 					// semaphore needed on start and creation of thread. 

    // public data 
    wait_queue_head_t queue;// queue thread is waiting on. Gets initialized by init_kthread, can be used by thread itself.
    int terminate; 			//flag to tell thread whether to die or not. When the thread receives a signal, it must check
				   			//the value of terminate and call exit_kthread and terminate if set.
    void *arg;				//additional data to pass to kernel thread
} kthread_t;

// start new kthread (called by creator) 
void start_kthread(void (*func)(kthread_t *), kthread_t *kthread);

// stop a running thread (called by "killer")
void stop_kthread(kthread_t *kthread);

// setup thread environment (called by new thread) 
void init_kthread(kthread_t *kthread, char *name);

// cleanup thread environment (called by thread upon receiving termination signal) 
void exit_kthread(kthread_t *kthread);

#endif 
