/*
Copyright (C) 2002  Sensoray Co., Inc.  This program is free software;
you can redistribute it and/or modify it under the terms of the GNU
General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

In addition, as a special exception, Sensoray Co., Inc., gives permission
to link the code of this program with the 'x12cnt.o' library (or with
modified versions of x12cnt.o that use the same license as 'x12cnt.o'),
and distribute linked combinations including the two. You must obey the
GNU General Public License in all respects for all of the code used other
than 'x12cnt.o'. If you modify this file, you may extend this exception
to your version of the file, but you are not obligated to do so. If you
do not wish to do so, delete this exception statement from your version.
*/

/* Deal with CONFIG_MODVERSIONS */
#include <linux/config.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

/* The necessary header files */

#include <linux/module.h>   /* Specifically, a module */
#include <asm/uaccess.h>    /* for get_user and put_user */
#include <linux/kernel.h>   /* We're doing kernel work */
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/wrapper.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/poll.h>
#include <linux/proc_fs.h>

/* Special 512 id registers */
#define WR512REG  0x0e
#define RD512REG  0x02

#define HZ_DIVIDER 50
#define VQ_HIGH_LEVEL    0xc000 //default values
#define VQ_LOW_LEVEL     0x4000
#define SLEEP_TIMEOUT (HZ / 2)
#define FRAME_TIMEOUT (HZ / 2)
#define START_TIMEOUT (HZ * 3)
#define VQ_TIMEOUT    (HZ * 2)

#include "kthread.h"
#include "x12gbl.h"

// mpeg data in/out wait queues
wait_queue_head_t mpeg_inq [MAX_CARDS];
wait_queue_head_t mpeg_outq[MAX_CARDS];

/* pointers to VPAR structures */
VPAR *pvpar[MAX_CARDS];

/* Decompression and compression flags */
/* Non-zero value if in operation      */
static int RunningRD[MAX_CARDS]; /* compression */
static int RunningWR[MAX_CARDS]; /* decompression */
static int HeaderMode[MAX_CARDS] = {0,};
static int HeaderPos[MAX_CARDS] = {0,};
static int rd_crashed_flag[MAX_CARDS] = {0,};
static int rd_crash_watch_en[MAX_CARDS] = {0,};
static int wr_crashed_flag[MAX_CARDS] = {0,};
static int wr_crash_watch_en[MAX_CARDS] = {0,};
static int video_frames_on[MAX_CARDS] = {0,};
static int video_frames_off[MAX_CARDS] = {0,};
static int CardPresent[MAX_CARDS] = {0,};

static unsigned short num_cards;
static unsigned short *StreamBuffer[MAX_CARDS];

/* insmod parameters */

static unsigned short io     [MAX_CARDS + 1] = {0,}; /* 512 io base address  */
                                                     /* for default 512      */
						     /* jumper settings set  */
						     /* this to 0xcc0        */

static unsigned short intype [MAX_CARDS + 1] = {0,}; /* Input type:          */
                                                     /*  0 - Composite,      */
						     /*  1 - S-Video         */

static unsigned short inno   [MAX_CARDS + 1] = {0,}; /* Input number (0 - 4) */
						     /* Default is 0         */

static unsigned short system [MAX_CARDS + 1] = {0,}; /* Video system:        */
                                                     /* 0 - Default (NTSC)   */
                                                     /* 1 - PAL              */
						     /* 2 - NTSC             */

static unsigned long  bitrate[MAX_CARDS + 1] = {0,}; /* Bit rate:            */
                                                     /* 0 - Default 4MBits/s */

static unsigned short size   [MAX_CARDS + 1] = {0,}; /* Picture size:        */
                                                     /* 0 - Default          */
						     /*     D1, MPEG-2,      */
						     /*     NTSC-704x480     */
						     /*     PAL-704x576      */
                                                     /* 1 - 1/2 D1, MPEG-2,  */
						     /*     NTSC-352x480,    */
						     /*     PAL-352x576      */
			                             /* 2 - 2/3 D1, MPEG-2,  */
						     /*     NTSC-480x480,    */
						     /*     PAL-480x576      */
			                             /* 3 - CIF, MPEG-1,     */
						     /*     NTSC-352x240,    */
						     /*     PAL-352x288      */

static unsigned short vbron  [MAX_CARDS + 1] = {0,}; /* Variable bit rate    */
                                                     /* compession:          */
						     /* 1 - Enabled          */
						     /* 0 - Disabled(default)*/


MODULE_PARM(io,        "1-" __MODULE_STRING(MAX_CARDS) "h");
MODULE_PARM(system,    "1-" __MODULE_STRING(MAX_CARDS) "h");
MODULE_PARM(intype,    "1-" __MODULE_STRING(MAX_CARDS) "h");
MODULE_PARM(inno,      "1-" __MODULE_STRING(MAX_CARDS) "h");
MODULE_PARM(size,      "1-" __MODULE_STRING(MAX_CARDS) "h");
MODULE_PARM(vbron,     "1-" __MODULE_STRING(MAX_CARDS) "h");
MODULE_PARM(bitrate,   "1-" __MODULE_STRING(MAX_CARDS) "i");

// VQ thread stuff...
kthread_t VQSize;
static void VQSize_thread(kthread_t *kthread); // VQSize thread for blocked operations

/* ioctl method - the 'misc.' function of the driver*/
static int x12_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long val)
{
	int CardId = MINOR (inode -> i_rdev);
	int retval = 0;

  	switch(cmd)
	{
		/* Resets the 512 */
		case X12_IOC_RESET:
            RunningWR[CardId] = 0;
            RunningRD[CardId] = 0;
            HeaderPos[CardId] = 0;
            Reset(CardId);
            break;

		/* Select video input, i.e. 0-4*/
        case X12_IOC_SET_INPUT:
            pvpar[CardId]->input = val;
            break;

    	        /* Sets video input type, i.e. SVIDEO (1) or CVIDEO (0)*/
		case X12_IOC_SET_INPUT_TYPE:
            pvpar[CardId]->inputtype = val;
            break;

		/* Set video system, i.e. NTSC (2) or PAL (1) */
		case X12_IOC_SET_SYSTEM:
            pvpar[CardId]->video_system = val;
            break;

		/* Set desired bit rate, typically 700,000 to 10,000,000*/
		case X12_IOC_SET_BITRATE:
            pvpar[CardId]->bit_rate = (double) val;
            bitrate[CardId] = val;
            break;

		/* Set picture size and compression system */
		/* D1 MPEG-2 (0), 1/2 D1 MPEG-2 (1),       */
		/* 2/3 D1 MPEG-2 (2), or CIF MPEG-1 (3)    */
		case X12_IOC_SET_PICSIZE:
            pvpar[CardId]->pic_size = (long) val;
      		break;

		/* Sets number of frames in Group of Pictures */
		case X12_IOC_SET_M:
            pvpar[CardId]->M = val;
      		break;

		/* Sets distance between I/P frames */
		case X12_IOC_SET_N:
            pvpar[CardId]->N = val;
      		break;

		/* Special header mode -- only works with Sensoray Model 301 CPU*/
		case X12_IOC_SET_HEADER_MODE:
            HeaderMode[CardId] = val;
      		break;

		/* Enables (1)/disables (0) variable bit rate compression */
		case X12_IOC_SET_VBR:
    		pvpar[CardId]->VBR_on = val;
    		break;

		/* Reads state of GPIO pins */
		case X12_IOC_READ_GPIO:
			retval = (~ReadGPIO(CardId)) & 7;//active low inputs
			break;                     //when input pulled down
						   //bit is set to 1
		/* Sets state of GPIO pins */
		case X12_IOC_WRITE_GPIO:
			WriteGPIO(CardId, 0, val); //active low outputs
            break;                     //when bit is set to 1
			                           //output goes low
   		/* Returns compression frame count */
		case X12_IOC_GET_FRAMECOUNT:
			retval = GetFrameCount(CardId);
      		break;

		/* Returns decompression frame count */
		case X12_IOC_GET_DISPLAYFRAMECOUNT:
			retval = GetDisplayFrameCount(CardId);
      		break;

		/* Returns number of dropped frames during compression*/
		case X12_IOC_GET_DROPPEDFRAMES:
			retval = GetDroppedFrames(CardId);
      		break;

        /* Returns amount of data in video queue */
   		case X12_IOC_GET_VQSIZE:
			retval = GetVQSize(CardId);
      		break;

        /* Enables frame rate reduction */
   		case X12_IOC_SET_REDUCED_FRAME_RATE:
			video_frames_off[CardId] = (val >> 8) & 0xff; // high byte
			video_frames_on[CardId]  = val & 0xff;        // low byte
      		break;

		/* If non of the above... */
   		default:
  			return -EINVAL;
  			break;
  	}
	return retval;
}

/* Read method - Attempts to read "count" bytes of MPEG data from the
SM2210 output fifo. The actual ammount read is returned. If no data
is available EAGAIN is returned unless blocking has been enabled. In
which case this function will wait until at least one byte of data is
made available. If this is the first read the encoder is initialized
and started.*/
static ssize_t x12_read(struct file *file, char *buffer, size_t nbytes, loff_t *ppos)
{
    struct inode *inode = file->f_dentry->d_inode;
	int CardId = MINOR (inode -> i_rdev);
	int npos, nlen;
	long BytesRead = 0;

	// if 512 has crashed return with fault
	if(rd_crashed_flag[CardId])
		return -EIO;

	/* if this is the first time start the encoder */
	if(!RunningRD[CardId])
	{
		if(StartEncode(CardId) < 0)
			return -EFAULT;

		// enable thread polling of VQ size
		RunningRD[CardId] = -1;
		RunningWR[CardId] = 0;

		// wait for data before proceeding
		interruptible_sleep_on_timeout(&mpeg_inq[CardId], START_TIMEOUT);

		rd_crash_watch_en[CardId] = 1;  // enable 512 crash watch
	}

	// check that nbytes is no too big.
	if(nbytes > MAX_BUF) nbytes = MAX_BUF;

	// Special header mode only available on Sensoray 301
	if(HeaderMode[CardId])
	{
		// get current buffer pointer
		npos = HeaderPos[CardId];
		while(ReadNextHeader(CardId, StreamBuffer[CardId], nbytes, &nlen) <= 0)
		{
			if(nlen > 0) // add data to user buffer
			{
				if(copy_to_user(&buffer[npos], StreamBuffer[CardId], nlen) != 0)
					return -EFAULT;

				// update buffer pointers
				npos += nlen;
				HeaderPos[CardId] = npos;
			}

			// if non-blocking return
			if(file->f_flags & O_NONBLOCK)
				return -EAGAIN;

			// in blocking mode, wait for more data
			interruptible_sleep_on_timeout(&mpeg_inq[CardId],SLEEP_TIMEOUT);

			// if 512 crashed return with fault
			if(rd_crashed_flag[CardId])
				return -EIO;

		}
		// mpeg block is complete. Add new data to user buffer
		HeaderPos[CardId] = 0;
		BytesRead = npos + nlen;
		if(copy_to_user(&buffer[npos], StreamBuffer[CardId], nlen) != 0)
			return -EFAULT;
	}
	else // this is the standard read
	{
		while((BytesRead = ReadStream(CardId, StreamBuffer[CardId], nbytes)) == 0)
		{
			// if non-blocking return
			if(file->f_flags & O_NONBLOCK)
				return -EAGAIN;

			// in blocking mode, wait for more data
			interruptible_sleep_on_timeout(&mpeg_inq[CardId], SLEEP_TIMEOUT);

			// if 512 crashed return with fault
			if(rd_crashed_flag[CardId])
				return -EIO;
		}

		if(copy_to_user(buffer, StreamBuffer[CardId], BytesRead) != 0)
			return -EFAULT;
	}

	return BytesRead;
}

/* Write method - Attempts to write "count" bytes of MPEG data to the
SM2210 input fifo. The actual ammount written is returned. If this
is the first write the decoder is initialized and started.*/
static ssize_t x12_write (struct file *file, const char *buffer, size_t count, loff_t *ppos)
{
    struct inode *inode = file->f_dentry->d_inode;
	int CardId = MINOR (inode -> i_rdev);
	unsigned long BytesWritten = 0;

//  printk ("sx12(%d) write called\n",CardId);
	// if 512 has crashed return with fault
	if(wr_crashed_flag[CardId])
		return -EIO;

	/* if this is the first time start the decoder */
	if(!RunningWR[CardId])
	{
		if(StartDecode(CardId) < 0)
			return -EFAULT;

		RunningWR[CardId] = -1;
		RunningRD[CardId] = 0;

		wr_crash_watch_en[CardId] = 1;  // enable 512 crash watch
	}

	if(count > MAX_BUF) count = MAX_BUF;

	if(copy_from_user(StreamBuffer[CardId], buffer, count) != 0)
    	        return -EFAULT;

	while(1)
	{
		if(GetVQSize(CardId) < VQ_HIGH_LEVEL)
		{
			BytesWritten = WriteStream(CardId, StreamBuffer[CardId], count);
			break;
		}
		else
		{
			if(file->f_flags & O_NONBLOCK)
				return -EAGAIN;

			interruptible_sleep_on_timeout(&mpeg_outq[CardId], SLEEP_TIMEOUT);

			// if 512 crashed return with fault
			if(wr_crashed_flag[CardId])
				return -EIO;
		}
	}
	return BytesWritten;
}

/* Open method - check that the minor device is valid. If not, return
   ENODEV.  Increment usage counter (many processes may use this device
   simultaneously), allocate buffer. */
static int x12_open (struct inode *inode, struct file *file)
{
	int CardId = MINOR (inode -> i_rdev);

//	printk("open sx12(%d)\n", CardId);

	if (CardId > num_cards || CardId < 0 || CardPresent[CardId] == 0 )
	{
		printk ("Error: invalid minor device number\n");
		return -ENODEV;
	}
   	StreamBuffer[CardId] = (unsigned short *) vmalloc(MAX_BUF);
	if( StreamBuffer[CardId] == 0)
	{
		printk ("Error: buffer allocation failed\n");
		return -ENOMEM;

	}
	RunningWR[CardId] = 0;
	RunningRD[CardId] = 0;
	HeaderPos[CardId] = 0;
	HeaderMode[CardId] = 0;
	rd_crashed_flag[CardId] = 0;
	rd_crash_watch_en[CardId] = 0;
	wr_crashed_flag[CardId] = 0;
	wr_crash_watch_en[CardId] = 0;
    video_frames_on[CardId] = 0;
    video_frames_off[CardId] = 0;

	Reset(CardId);
	MOD_INC_USE_COUNT;

//	printk ("sx12(%d) open. Buffer address: %lx\n", CardId, (unsigned long)StreamBuffer[CardId]);

	return 0;
}

/* Close method - release buffer and decrement the usage counters. */
static int x12_release(struct inode *inode, struct file *file)
{
	int CardId = MINOR (inode -> i_rdev);

    VideoInputControl(CardId, 1);		// turn video back on

//	printk("close sx12(%d)\n", CardId);

    RunningWR[CardId] = 0;
	RunningRD[CardId] = 0;
	HeaderPos[CardId] = 0;
	HeaderMode[CardId] = 0;
	rd_crashed_flag[CardId] = 0;
	rd_crash_watch_en[CardId] = 0;
	wr_crashed_flag[CardId] = 0;
	wr_crash_watch_en[CardId] = 0;
	video_frames_on[CardId] = 0;
    video_frames_off[CardId] = 0;


	vfree(StreamBuffer[CardId]);
	Reset(CardId);

	MOD_DEC_USE_COUNT;

    return 0;
}

/* The poll method */
unsigned int x12_poll(struct file *file, poll_table *wait)
{
    struct inode *inode = file->f_dentry->d_inode;
	int CardId = MINOR (inode -> i_rdev);
	unsigned int mask = 0;

	// if the 512 has crashed return with fault
	if(rd_crashed_flag[CardId])
		return -EIO;

	// if the encoder or decoder has not been started
	// return as if all is ok! No blocking operations
	// may not have any data and must be checked because
	// of this fix.
	if(RunningRD[CardId] == 0 && RunningWR[CardId] == 0)
		return POLLIN | POLLRDNORM | POLLOUT | POLLWRNORM;

	poll_wait(file, &mpeg_outq[CardId], wait);
	poll_wait(file, &mpeg_inq [CardId], wait);

	if (GetVQSize(CardId) > VQ_LOW_LEVEL ) mask |= POLLIN  | POLLRDNORM; /* readable */
	if (GetVQSize(CardId) < VQ_HIGH_LEVEL) mask |= POLLOUT | POLLWRNORM; /* writable */

	return mask;
}



int x12_read_procmem ( char *buf, char **start, off_t offset, int count, int *eof, void *data )
{
    int i, len = 0;

    int limit = count - 80; /* Don't print more than this */

	for (i = 0; i < num_cards && len <= limit; i++)
	{
		if ( CardPresent[i] )
		{
			if ( RunningRD[i] )
	        	len += sprintf( buf + len, "Card %d: ENCODE  vq %ld, fr %ld\n", i, GetVQSize(i), GetFrameCount(i) );
			else if ( RunningWR[i] )
	        	len += sprintf( buf + len, "Card %d: DECODE  vq %ld, fr %ld\n", i, GetVQSize(i), GetFrameCount(i) );
			else
	        	len += sprintf( buf + len, "Card %d: STANDBY vq %ld, fr %ld\n", i, GetVQSize(i), GetFrameCount(i) );
		}
    }
    *eof = 1;
    return len;
}



/* The access methods are placed in a table of valid file operations. This
   driver supports open, release (close), read, write and ioctl. Use the new
   gcc extension syntax. */

struct file_operations x12_fops = {
	read:		x12_read,
	write:		x12_write,
	poll:		x12_poll,
	ioctl:		x12_ioctl,
	open:		x12_open,
	release:	x12_release
};




/* This is called when the module is loaded. It checks and allocates major
   number, base addresses, number of 512s, etc.  */
int init_module (void)
{
	int result;
	int i;
	int AnyCards = 0;

	printk ("\nInstalling 512 driver...     (Build %d)\n", BUILD);


	/* Counting number of 512s */
	for(num_cards = 0; io[num_cards] != 0 && num_cards < MAX_CARDS; num_cards++);

//	printk("numcards = %d\n", num_cards);

	/*if there are no 512s quit */
	if(num_cards == 0)
	{
		printk ("Missing 'io=' module parameter.\n");
		printk ("512 driver installation failed!\n\n");
		return -1;
	}

	/*if there too many 512s quit */
	if(num_cards > MAX_CARDS)
	{
		printk ("Too many 512s. Reduce the number of 'io=' module parameters.\n");
		printk ("512 driver installation failed!\n\n");
		return -1;
	}

	/*Register the device*/
	if ((result = register_chrdev (X12_MAJOR, "sx12", &x12_fops)))
	{
		printk ("Unable to get major number %i for 512.\n", X12_MAJOR);
		printk ("512 driver installation failed!\n\n");
		return result;
	}

        /*Initialize each 512*/
	for (i = 0; i < num_cards; i++)
	{
	        /* set the default video system - NTSC */
		system[i] = system[i] == 0 ? NTSC : system[i];

		printk("512(%d) io=%#x",i,io[i]);

		if ((result = check_region (io[i], 32)) < 0)
		{
			printk ("\n512(%d) io address space is already occupied: %#x.\n", i,io[i]);
			unregister_chrdev (X12_MAJOR, "sx12");
			printk ("512 driver installation failed!\n\n");
			return result;
		}
                /******************************** VERY IMPORTANT ***************************************/
                /* The arguments &jiffies, printk, schedule, and HZ must always appear as shown below. */
                /***************************************************************************************/
		if((pvpar[i] = Initialize(i, io[i], system[i], intype[i],inno[i], &jiffies,
		    printk, schedule, HZ, VQ_HIGH_LEVEL, VQ_LOW_LEVEL)) == 0)
		{
       		printk("\n512(%d) not installed at io address %#x.\n", i, io[i]);
			CardPresent[i] = 0;
			continue;
		}
		else
		{
			CardPresent[i] = 1;
			AnyCards = 1;

			request_region (io[i], 32, "sx12");

			/* set default bitrate */
			if(bitrate[i] == 0)
					bitrate[i] = 4000000L;

			pvpar[i]->bit_rate = bitrate[i];
					pvpar[i]->pic_size = size[i];
			pvpar[i]->VBR_on   = vbron[i];

			printk(",system=%d,inno=%d,intype=%d,bitrate=%ld,size=%d,vbron=%d.\n"
					,system[i],inno[i],intype[i],bitrate[i], size[i],vbron[i]);
			printk ("512(%d) at %#x initialised.\n",i, io[i]);
		}
	}
	if(!AnyCards)
	{
		unregister_chrdev (X12_MAJOR, "sx12");
		printk ("512 driver installation failed!\n\n");
		return -1;
	}

	start_kthread(VQSize_thread, &VQSize);

	printk ("512 driver successfuly installed!\n\n");

	create_proc_read_entry("sx12",
	                  		0    /* default mode */,
                       		NULL /* parent dir */,
                       		x12_read_procmem,
                       		NULL /* client data */);

	return 0;
}

/* This is called when the module is unloaded*/
void cleanup_module (void)
{
	int i;

	stop_kthread(&VQSize);

	/* reset each 512*/
	for (i = 0; i < num_cards; i++)
	{
		if(CardPresent[i])
		{
			Reset(i);
			release_region (io[i], 32);
			CardPresent[i] = 0;
		}
	}
	/* unregister device */
	unregister_chrdev (X12_MAJOR, "sx12");

	printk ("512 driver removed!\n");
}


/* this is the thread function that we are executing */

static void VQSize_thread(kthread_t *kthread)
{
    unsigned long frame_count, old_frame_count[MAX_CARDS];
    unsigned long old_jiffies[MAX_CARDS];
    int video_on_off_frame_count[MAX_CARDS];
    int video_input_on[MAX_CARDS];
    int new_FID;
    int old_FID[MAX_CARDS];
	int vq, i, old_vq[MAX_CARDS];

    init_kthread(kthread, "sx12_thread");	// setup the thread environment

	for(i = 0; i < num_cards; i++)
	{
		if(CardPresent[i])
		{
			// initializing the wait queues
			init_waitqueue_head(&mpeg_inq [i]);
			init_waitqueue_head(&mpeg_outq[i]);

			// video is initially on
			video_input_on[i] = 1;
			video_on_off_frame_count[i] = 0;
		}
	}

    while(1)    // an endless loop in which we are doing our work
    {
		interruptible_sleep_on_timeout(&kthread->queue, HZ/ HZ_DIVIDER);// fall asleep
		mb();// We need to do a memory barrier here to be sure that the flags are visible on all CPUs.
		if (kthread->terminate)
		{
			break;  // we received a request to terminate ourself
		}

		for(i = 0; i < num_cards; i++)
		{
			if(CardPresent[i])
			{
				// get video queue size
				vq = GetVQSize(i);

				// if reading (encoding) ...
				if(RunningRD[i])
				{
					// if there is mpeg data available, wake up wait queue
					if(vq > VQ_LOW_LEVEL)
					{
						wake_up_interruptible(&mpeg_inq[i]);
					}

					// if crash watch enabled check for non-changing frame number
					if(rd_crash_watch_en[i])
					{
						frame_count = GetFrameCount(i);

						if (frame_count == old_frame_count[i])
						{
							if((jiffies - old_jiffies[i]) > FRAME_TIMEOUT)
							{
								rd_crashed_flag[i] = 1;
							}
						}
						else
						{
							old_jiffies[i] = jiffies;
							old_frame_count[i] = frame_count;
						}
					}
					else
					{
						old_frame_count[i] = GetFrameCount(i);
						old_jiffies[i] = jiffies;
					}
				}

				// if writing (decoding) ...
				if(RunningWR[i])
				{
					// if there is space to add more mpeg data, wake up wait queue
					if(vq < VQ_HIGH_LEVEL)
					{
						wake_up_interruptible(&mpeg_outq[i]);
					}
					// if crash watch enabled check for non-changing VQ size
					if(wr_crash_watch_en[i])
					{
						if (vq == old_vq[i])
						{
							if((jiffies - old_jiffies[i]) > VQ_TIMEOUT)
							{
								wr_crashed_flag[i] = 1;
							}
						}
						else
						{
							old_jiffies[i] = jiffies;
							old_vq[i] = vq;
						}
					}
					else
					{
						old_vq[i] = vq;
						old_jiffies[i] = jiffies;
					}
				}

				// frame rate reduction
				new_FID = GetInputVideoFID(i);
				if(video_frames_off[i] && video_frames_on[i])
				{
					// count frames on positive FID transition
					if(new_FID && !old_FID[i])
					{
						video_on_off_frame_count[i]++;
					}

					// if on frame count reached while video input is on, turn off video input
					if(video_input_on[i] && (video_on_off_frame_count[i] == video_frames_on[i]))
					{
						video_input_on[i] = 0;			// clear video input on flag
						video_on_off_frame_count[i] = 0;// reset frame counter
						VideoInputControl(i, 0);		// turn video off
					}
					// if off frame count reached while video input is off, turn on video input
					else if (!video_input_on[i] && (video_on_off_frame_count[i] == video_frames_off[i]))
					{
						video_input_on[i] = 1;			// set video input on flag
						video_on_off_frame_count[i] = 0;// reset frame counter
						VideoInputControl(i, 1);		// turn video back on
					}
				}
				old_FID[i] = new_FID;
			}
		}
	}
    exit_kthread(kthread); //cleanup the thread, leave
}

// Set driver license.
MODULE_LICENSE( "GPL" );





